/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

/**
 * @file at_command_mode/application/include/application.h
 * @brief Header file for the common application parts
 * @details This header files contains the common application parts that are
 * shared between the peripheral and central demo.
 *
 * Please note that the central
 * demo is not fully functional yet, because the AT command firmware is
 * missing the GATT client and the SPPoverBLE central implementations.
 **/

#ifndef __APPLICATION_H__
#define __APPLICATION_H__

#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <windows.h>

/*! Definition of the maximum size of a ringbuffer instance */
#define RING_SIZE 64
/*! Typedef for a position in the ringbuffer. Should be the "native" type of the processor */
typedef uint32_t ring_pos_t;

/*! Structure for a ringbuffer */
struct ringbuffer {
  volatile ring_pos_t ring_head;         /*!< "Pointer" to the head of the rinbuffer */
  volatile ring_pos_t ring_tail;         /*!< "Pointer" to the tail of the ringbuffer */
  volatile uint8_t ring_data[RING_SIZE]; /*!< Data buffer for the ringbuffer */
};

/**
 * @brief Function that flushes a ringbuffer
 *
 * @param[in] rb Pointer to the ringbuffer
 *
 * @return 0 if success, -1 if error
 */
void flush_rb(struct ringbuffer *rb);

/**
 * @brief Function that adds one element to a ringbuffer
 *
 * @param[in] rb Pointer to the ringbuffer
 * @param[in] c Element to be added
 *
 * @return 0 if success, -1 if error
 */
int add_to_rb(struct ringbuffer *rb, uint8_t c);

/**
 * @brief Function that removes one element from a ringbuffer
 *
 * @param[in] rb Pointer to the ringbuffer
 * @param[in] c Pointer where the removed element is stored
 *
 * @return 0 if success, -1 if error or ringbuffer is empty
 */
int remove_from_rb(struct ringbuffer *rb, uint8_t *c);

#ifndef BD_ADDR_FORMAT
/*! Definition of a format string for printing a Bluetooth device address */
#define BD_ADDR_FORMAT "%02x:%02x:%02x:%02x:%02x:%02x"
#endif

#ifndef BD_ADDR_ARG
/*! Definition of an argument list for printing a Bluetooth device address */
#define BD_ADDR_ARG(x) x[0], x[1], x[2], x[3], x[4], x[5]
#endif

/*! Definition of the maximum length of the line input buffer. It must be large enough
to hold the longest return line from th AT command firmware */
#define MAX_BUF_SIZE 1024
/*! Declaration of a send and receive buffer for AT commands, reponses and events */
extern uint8_t buf[MAX_BUF_SIZE];

/*! Definition of the "comma" that delimits arguments in an AT command */
#define COMMA ','
/*! Definition of the "end of command" that ends an AT command */
#define END '\r', '\n'

/**
 * @brief Function that initialises the application. For the peripheral
 * implementation this includes starting the advertising.
 *
 * @param[in] device_name Pointer to a device name to be used
 *
 * @return 0 if success, -1 if error
 */
int application_init(const char *device_name);

/**
 * @brief Function that handles regular application processing. It is called whenever the
 * application has been woken up, either due to Bluetooth handling, UART handling or application
 * handling. For the peripheral demo, keypresses to be sent to the remote device, for example, are collected in a
 * ringbuffer. When this function is called, the application should check if data can be send and if so, initiate the
 * send operation.
 *
 * @return 0 if success, -1 if error
 */
int application_process(void);

/**
 * @brief Function that gracefully shuts down the application. If a connection exists
 * it will be terminated first.
 */
void application_shutdown(void);

/**
 * @brief Function that handles any keyboard input. For the peripheral demo, keypresses to be sent to the remote device
 * are collected in a ringbuffer for later processing in application_process().
 *
 * @param[in] key ASCII value of the key that was pressed
 *
 * @return 0 if success, -1 if error
 */
int application_handle_keyboard_input(int key);

/**
 * @brief Function that handles any AT event that was received.
 *
 * @param[in] ascii Pointer to the ASCII part of the event that was received
 * @param[in] ascii_length Length of the ASCII part
 * @param[in] binary Pointer to the binary part or argument part of the event that was received
 * @param[in] binary_length Length of the binary part or argument part
 *
 * @return 0 if success, -1 if error
 */
int application_handle_at_event(const uint8_t *ascii, const int ascii_length, const uint8_t *binary, const int binary_length);

/**
 * @brief Function that sends an AT command.
 *
 * @param[in] cmd Pointer to AT command to be send
 * @param[in] len Length of AT command
 *
 * @return 0 if success, -1 if error
 */
int at_transfer_command(const uint8_t *cmd, unsigned int len);

/**
 * @brief Function that sends an AT command, expecting a status code from the +OK response
 *
 * @param[in] cmd Pointer to AT command to be send
 * @param[in] len Length of AT command
 * @param[in] cmd Pointer to variable that will hold the status code
 *
 * @return 0 if success, -1 if error
 */
int at_transfer_command_status(const uint8_t *cmd, unsigned int len, uint8_t *status_code);

    /**
 * @brief Function that tries to receive a line from the AT command firmware
 *
 * @param[in] buf Pointer to a buffer to hold the received line
 * @param[in] max_length Length of the buffer or maximum length of the line to be received
 * @param[out] length Pointer to an integer that will receive the number of bytes that have been received
 * @param[in] timeout Maximum time to wait for the line to be received
 *
 * @return 0 if success, -1 if error or timeout
 */
int at_receive_line(uint8_t *buf, unsigned int max_length, unsigned int *length, int timeout);

/**
 * @brief Function that sends the AT LEDISCON" command to terminate an existing connection
 *
 * @param[in] bd_address Pointer to the Bluetooth address of the device to disconnect from
 *
 * @return 0 if success, -1 if error
 */
int at_send_lediscon(uint8_t *bd_address);

/*! Definition for the maximum MTU size */
#define SER_ATT_MTU_SIZE_MAX 247

/*! Structure for the application data that is absolutely necessary */
struct application {
  uint8_t buf[SER_ATT_MTU_SIZE_MAX]; /*!< Intermediate buffer used to assemble data to be send and to process data that was received */
  bool is_connected;                 /*!< Indicates if a connection is present or not */
  uint16_t connection_id;            /*!< Connection identifier for an existing connection */
  uint8_t bd_address[6];             /*!< Bluetooth device address of the connected device */
  uint16_t mtu_size;                 /*!< MTU size that was negotiated for the existing connection */
  bool is_indication_enabled;        /*!< Indicates if indications have been enabled (peripheral processing only) */
  bool is_sending;                   /*!< Indicates if the device is currently sending or not */
  struct ringbuffer in;              /*!< Ringbuffer used as intermediate storage for keypresses */
};

/*! Declaration of the main application structure */
extern struct application application;

#ifdef DEBUG
/*! Definition of the debug printf macro that will output some statistical information as well as the function name
 * before the debug message itself */
#define dprintf(...)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  \
  do {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                \
    printf("[%d|%d|%d] %s(): ", application.is_connected, application.is_indication_enabled, application.mtu_size, __func__);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         \
    printf(__VA_ARGS__);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              \
  } while (0)
#else
/*! Empty definition of the debug printf macro when debugging is not enabled */
#define dprintf(...)
#endif

/*! Definition of the warning printf macro that will output some statistical information as well as the function name
 * before the warning message itself */
#define warning(...)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  \
  do {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                \
    printf("[%d|%d|%d] %s(): ", application.is_connected, application.is_indication_enabled, application.mtu_size, __func__);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                         \
    printf(__VA_ARGS__);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              \
  } while (0)

/*! Definition of an unconditional output printf macro that will output some information */
#define xprintf(...)                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                  \
  do {                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                \
    printf(__VA_ARGS__);                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                                              \
  } while (0)

#endif