/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

/**
 * @file at_command_mode/application/include/uart.h
 * @brief Header file for a simple UART abstraction layer
 * @details This header files contains a simple UART abstraction layer to
 * send data and receive whole "lines" from the AT command firmware.
 **/

#ifndef __UART_H__
#define __UART_H__

#include <stdint.h>
#include <windows.h>

/**
 * @brief Function that will open a UART port that is determined by the following logic: all available FTDI devices will
 * be scanned. If one (and only one) UART port is not already opened, then this UART port will be opened.
 *
 * @param[in] handle Pointer to a handle that will be returned and can be used for further processing like waiting
 * @param[in] flow_control Indicates if flow control to that device shall be enabled or not
 *
 * @return 0 if success, -1 if error
 */
int uart_init(HANDLE *handle, BOOL flow_control);

/**
 * @brief Function that will change the baudrate of the UART port. The UART port must already be initialized by
 * uart_init()
 *
 * @param[in] baudrate New baudrate to be used
 *
 * @return 0 if success, -1 if error
 */
int uart_set_baudrate(int baudrate);

/**
 * @brief Function that send data to the UART.
 *
 * @param[in] buf Pointer to a buffer memory that contains the data to be send
 * @param[in] length Length of the data to be send
 *
 * @return 0 if success, -1 if error
 */
int uart_send_bytes(const uint8_t *buf, unsigned int length);

/**
 * @brief Function that will receive a whole "line" from the UART. Because each "event" from the
 * AT command firmware is prefixed with the length of the data to expect, this function can easily
 * determine when a "line" is fully received or not.
 *
 * @param[in] buf Pointer to a buffer to hold the received line
 * @param[in] max_length Length of the buffer or maximum length of the line to be received
 * @param[out] length Pointer to an integer that will receive the number of bytes that have been received
 * @param[in] timeout Maximum time to wait for the line to be received
 *
 * @return 0 if success, -1 if error
 */
int uart_receive_line(uint8_t *buf, unsigned int max_length, unsigned int *length, int timeout);

#endif