/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

#include "application.h"

#define NAME 'P', 'A', 'N', '-', 'U', 'S', 'B', '-', 'C', 0x00

static uint16_t mtu = 23;

static uint8_t peer_bd_address[6];

static int at_gssrvind(const uint8_t *data, int len)
{
  uint16_t timeout = 1000;
  int ret;

  uint8_t at_gssrvind_preamble[] = {
      'A',
      'T',
      '+',
      'G',
      'S',
      'S',
      'R',
      'V',
      'I',
      'N',
      'D',
      ' ',
      'P',
      3,
      '+',
      'C',
      1,
      COMMA,
      0x00,
      0x01,
      COMMA, /* connection id */
      timeout >> 8,
      timeout & 0xff,
      COMMA, /* timeout */
  };

  uint8_t *ptr = buf;
  memcpy(ptr, at_gssrvind_preamble, sizeof(at_gssrvind_preamble));
  ptr += sizeof(at_gssrvind_preamble);
  memcpy(ptr, data, len);
  ptr += len;
  *ptr++ = '\r';
  *ptr++ = '\n';

  ret = at_transfer_command(buf, (unsigned int)(ptr - buf));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

static int at_lesppscan_enable(void)
{
  uint8_t command[] = {'A',  'T',   '+', 'L', 'E', 'S', 'P', 'P', 'S', 'C', 'A', 'N', ' ', /* command */
                       0x01, COMMA,                                                        /* enable flag */
                       0x00,                                                               /* filter policy */
                       END};
  int ret;

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

static int at_lesppscan_disable(void)
{
  uint8_t command[] = {'A',  'T', '+', 'L', 'E', 'S', 'P', 'P', 'S', 'C', 'A', 'N', ' ', /* command */
                       0x00,                                                             /* enable flag */
                       END};
  int ret;

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

static int at_lesppconn(uint8_t *bd_address)
{
  uint8_t command[] = {'A',  'T',   '+',  'L',  'E',  'S',  'P', 'P', 'C', 'O', 'N', 'N', ' ', /* command */
                       0x01, COMMA,                                                            /* flag: initiate connection */
                       0x00, COMMA,                                                            /* addr type: public */
                       0x00, 0x00,  0x00, 0x00, 0x00, 0x00,                                    /* bd address */
                       END};
  int ret;

  memcpy(&command[17], bd_address, 6);

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

int application_handle_at_event(const uint8_t *ascii, const int ascii_length, const uint8_t *binary, const int binary_length)
{
  int ret;

  ret = strcmp(ascii, "LE_CONN_COMPLETED_EVENT");
  if (!ret) {
    int offset = 0;
    uint8_t status = binary[offset++];
    offset++; /* comma */
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;

    application.is_connected = true;

    dprintf("LE_CONN_COMPLETED_EVENT: status %d, connection_id %04x\n", status, connection_id);
    return 0;
  }

  ret = strcmp(ascii, "LE_DISCONN_EVENT");
  if (!ret) {
    int offset = 0;
    uint8_t reason = binary[offset++];
    offset++; /* comma */
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;

    application.is_connected = false;

    dprintf("LE_DISCONN_EVENT: reason %d, connection_id %04x\n", reason, connection_id);
    return 0;
  }

  ret = strcmp(ascii, "LE_SPPSCAN_EVENT");
  if (!ret) {
    int offset = 0;
    uint8_t advertising_type = binary[offset++];
    offset++; /* comma */
    uint8_t peer_address_type = binary[offset++];
    offset++; /* comma */
    memcpy(&peer_bd_address[0], &binary[offset], sizeof(peer_bd_address));
    offset += sizeof(peer_bd_address) + 1;
#if 0
		uint8_t name_length = binary[offset++];
		uint8_t name[20];
		if (name_length > 0) {
			memcpy(&name[0], &binary[offset], name_length);
			offset += name_length + 1;
		}
		uint8_t bond_status = binary[offset++];
		uint8_t data_len = binary[offset++];
		uint8_t data[31];
		if (data_len > 0) {
			memcpy(&data[0], &binary[offset], data_len);
			offset += data_len + 1;
		}
#endif
    dprintf("LE_SPPSCAN_EVENT: bd_addr " BD_ADDR_FORMAT "\n", BD_ADDR_ARG(peer_bd_address));

    return 0;
  }

  dprintf("unhandled event '%s'\n", ascii);
  return 0;
}

int application_init(const char *device_name)
{
  dprintf("central\n");

  uint8_t init[] = {'A',
                    'T',
                    '+',
                    'L',
                    'E',
                    'S',
                    'P',
                    'P',
                    'I',
                    'N',
                    'I',
                    'T',
                    ' ', /* command */
                    0x00,
                    COMMA, /* authentication requirement */
                    0x00,
                    COMMA, /* uuid flag */
                    0x12,
                    0x34,
                    COMMA, /* appearance */
                    (mtu >> 8) & 0xff,
                    mtu & 0xff,
                    COMMA, /* mtu */
                    0x00,
                    COMMA, /* connection parameters */
                    0x00,  /* scan / advertise flag */
                    END};

  int ret;

  ret = at_transfer_command(init, sizeof(init));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  dprintf("ok\n");

  return 0;
}

int application_process(void)
{
  return 0;
}

int application_handle_keyboard_input(int key)
{
  int ret;

  dprintf("'%d'\n", key);

  if (key == '1') {

    ret = at_lesppscan_enable();
    if (ret) {
      dprintf("at_lesppscan_enable() failed\n");
      return -1;
    }

    return 0;
  }

  if (key == '2') {

    ret = at_lesppscan_disable();
    if (ret) {
      dprintf("at_lesppscan_disable() failed\n");
      return 0;
    }

    return 0;
  }

  if (key == '3') {

    ret = at_lesppconn(peer_bd_address);
    if (ret) {
      dprintf("at_lesppconn() failed\n");
      return 0;
    }

    return 0;
  }

  if (key == '4') {

    ret = at_send_lediscon(peer_bd_address);
    if (ret) {
      dprintf("at_send_lediscon() failed\n");
      return 0;
    }

    return 0;
  }

  if (key == '0') {
    const char *hello = "Hello";
    return at_gssrvind(hello, (int)strlen(hello));
  }

  return 0;
}

void application_shutdown(void)
{
  int ret;

  if (application.is_connected == false) {
    return;
  }

  ret = at_send_lediscon(application.bd_address);
  if (ret) {
    dprintf("at_send_lediscon() failed\n");
    return;
  }
}
