/****************************************************************************
 *   COPYRIGHT 2019 All rights reserved
 *   Panasonic Industrial Devices Europe GmbH
 ****************************************************************************/

#include "application.h"

static uint8_t appearance[] = {0x00, 0x00};
static uint8_t device_name[0x1e] = "PAN1762-USB";
static size_t device_name_len = 11;
static uint8_t manufacturer_name_string[] = "Panasonic";
static uint8_t model_number_string[] = "PAN 1762";
static uint8_t system_id[] = "ETU";

static int at_gssrvind(uint8_t service, uint8_t characteristic, uint16_t timeout, const uint8_t *data, int len)
{
  int ret;

  // clang-format off
  uint8_t at_gssrvind_preamble[] = {
    'A', 'T', '+', 'G', 'S', 'S', 'R', 'V', 'I', 'N', 'D', ' ',
    'P', service, '+', 'C', characteristic, COMMA,
    0x00, 0x01, COMMA, /* connection id */
    timeout >> 8, timeout & 0xff, COMMA, /* timeout */
    /* data is added dynamically below */
  };
  // clang-format on

  uint8_t *ptr = buf;
  memcpy(ptr, at_gssrvind_preamble, sizeof(at_gssrvind_preamble));
  ptr += sizeof(at_gssrvind_preamble);
  memcpy(ptr, data, len);
  ptr += len;
  *ptr++ = '\r';
  *ptr++ = '\n';

  dprintf("len %d\n", len);

  ret = at_transfer_command(buf, (unsigned int)(ptr - buf));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_lesppadv_start(void)
{
  // clang-format off
  uint8_t command_preamble[] = {
    'A',  'T',   '+', 'L', 'E', 'S', 'P', 'P', 'A', 'D', 'V', ' ', /* command */
    0x01, COMMA, /* enable flag */
    /* device name + end is added dynamically below */
  };
  // clang-format on
  int ret;

  dprintf("\n");

  uint8_t *ptr = buf;
  memcpy(ptr, command_preamble, sizeof(command_preamble));
  ptr += sizeof(command_preamble);
  memcpy(ptr, device_name, device_name_len);
  ptr += device_name_len;
  *ptr++ = '\r';
  *ptr++ = '\n';

  ret = at_transfer_command(buf, (unsigned int)(ptr - buf));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_lesppinit(uint8_t authentication_requirement, uint8_t appearance[2], uint16_t mtu)
{
  // clang-format off
  uint8_t command[] = {
    'A', 'T', '+', 'L', 'E', 'S', 'P', 'P', 'I', 'N', 'I', 'T', ' ', /* command */
    authentication_requirement, COMMA, /* authentication requirement */
    0x00, COMMA, /* uuid flag */
    appearance[0], appearance[1], COMMA, /* appearance */
    (mtu >> 8) & 0xff, mtu & 0xff, COMMA, /* mtu */
    0x00, COMMA, /* connection parameters */
    0x00,  /* scan / advertise flag */
    END
  };
  // clang-format off
  int ret;

  dprintf("\n");

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_secsetparam(void)
{
  // clang-format off
  uint8_t command[] = {
    'A', 'T', '+', 'S', 'E', 'C', 'S', 'E', 'T', 'P', 'A', 'R', 'A', 'M', ' ', /* command */
    0x00, COMMA, /* i/o capabilities: display only */
    (0x01 << 0) | (1 << 2), COMMA, /* auth parameters: bonding enabled, mitm enabled  */
    0x07, COMMA, /* initiator keys */
    0x07, COMMA, /* responder keys */
    0x10, COMMA, /* encryption key size */
    0x00, /* oob flag */
    END
  };
  // clang-format off
  int ret;

  dprintf("\n");

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_gswrupd(uint16_t connection_id)
{
  // clang-format off
  uint8_t at_gswrupd[] = {
    'A',   'T', '+', 'G', 'S', 'W', 'R', 'U', 'P', 'D', ' ',
    connection_id >> 8, connection_id & 0xff, COMMA, /* connection id */
    0, /* accept */
    END
  };
  // clang-format on
  int ret;

  dprintf("sending confirmation\n");

  ret = at_transfer_command(at_gswrupd, sizeof(at_gswrupd));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_gsupdval(uint8_t service, uint8_t characteristic, uint8_t element, uint8_t *data, size_t len)
{
  // clang-format off
  uint8_t at_gsupdval_preamble[] = {
      'A', 'T', '+', 'G',   'S', 'U', 'P', 'D', 'V', 'A', 'L', ' ',
      'P', service, '+',
      'C', characteristic, '+',
      'E', element, COMMA,
      /* data + end is missing and is added dynamically */
  };
  // clang-format on
  int ret;

  uint8_t *ptr = buf;
  memcpy(ptr, at_gsupdval_preamble, sizeof(at_gsupdval_preamble));
  ptr += sizeof(at_gsupdval_preamble);
  memcpy(ptr, data, len);
  ptr += len;
  *ptr++ = '\r';
  *ptr++ = '\n';

  dprintf("updating data with len %zd for service %d, characteristic %d, element %d\n", len, service, characteristic, element);

  ret = at_transfer_command(buf, (unsigned int)(ptr - buf));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_gsrdupd(uint8_t service, uint8_t characteristic, uint8_t element)
{
  // clang-format off
  uint8_t at_gsrdupd[] = {
    'A', 'T', '+', 'G', 'S', 'R', 'D', 'U', 'P', 'D', ' ',
    'P', service, '+',
    'C', characteristic, '+',
    'E', element, COMMA,
    application.connection_id >> 8, application.connection_id & 0xff, COMMA, /* connection id */
    0, COMMA, /* update flag */
    0, /* accept value */ 
    END};
  // clang-format on
  int ret;

  dprintf("sending gsrdupd\n");

  ret = at_transfer_command(at_gsrdupd, sizeof(at_gsrdupd));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_secbondaccept(uint16_t connection_id)
{
  // clang-format off
  uint8_t command[] = {
    'A', 'T', '+', 'S', 'E', 'C', 'B', 'O', 'N', 'D', 'A', 'C', 'C', 'E', 'P', 'T', ' ',
    connection_id >> 8, connection_id & 0xff, COMMA, /* connection id */
    0, /* accept */
    END
  };
  // clang-format on
  int ret;

  dprintf("sending at_secbondaccept\n");

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

static int at_seccfmpkey(uint16_t connection_id)
{
  // clang-format off
  uint8_t command[] = {
    'A', 'T', '+', 'S', 'E', 'C', 'C', 'F', 'M', 'P', 'K', 'E', 'Y', ' ',
    connection_id >> 8, connection_id & 0xff, COMMA, /* connection id */
    0x12, 0x34, 0x56, /* key_value: 123456 */
    END
  };
  // clang-format on
  int ret;

  dprintf("sending at_seccfmpkey\n");

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  return 0;
}

int at_leresolvaddr(uint8_t *bd_address)
{
  uint8_t command[] = {'A',  'T',  '+',  'L',  'E',  'R',  'E', 'S', 'O', 'L', 'V', 'A', 'D', 'D', 'R', ' ', /* command */
                       0x00, 0x00, 0x00, 0x00, 0x00, 0x00,                                                   /* bd address */
                       END};
  uint8_t resolved_bd_address[6];
  int ret;

  memcpy(&command[16], bd_address, 6);

  ret = at_transfer_command(command, sizeof(command));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  unsigned int length;

  ret = at_receive_line(buf, sizeof(buf), &length, 1000);
  if (ret) {
    dprintf("at_receive_line() failed\n");
    return -1;
  }

  uint8_t *ptr = buf;

  static uint8_t response[] = {'+', 'L', 'E', 'R', 'E', 'S', 'O', 'L', 'V', 'A', 'D', 'D', 'R', '='};

  ret = memcmp(ptr, response, sizeof(response));
  if (ret) {
    dprintf("did not receive expected response\n");
    return -1;
  }

  ptr += sizeof(response);

  int offset = 0;
  memcpy(resolved_bd_address, &ptr[offset], 6);
  offset += 6;
  /* ignore the rest */

  xprintf("bd_addr " BD_ADDR_FORMAT " -> " BD_ADDR_FORMAT "\n", BD_ADDR_ARG(bd_address), BD_ADDR_ARG(resolved_bd_address));
  return 0;
}

static int at_leconinfo(uint16_t connection_id, uint8_t *bd_address)
{
  // clang-format off
  uint8_t at_leconinfo[] = {
    'A', 'T', '+', 'L', 'E', 'C', 'O', 'N', 'I', 'N', 'F', 'O', ' ',
    connection_id >> 8, connection_id & 0xff, /* connection id */
    END
  };
  // clang-format on
  int ret;

  dprintf("retrieve connection information\n");

  ret = at_transfer_command(at_leconinfo, sizeof(at_leconinfo));
  if (ret) {
    dprintf("at_transfer_command() failed\n");
    return -1;
  }

  unsigned int length;

  ret = at_receive_line(buf, sizeof(buf), &length, 1000);
  if (ret) {
    dprintf("at_receive_line() failed\n");
    return -1;
  }

  uint8_t *ptr = buf;

  static uint8_t leconinfo[] = {'+', 'L', 'E', 'C', 'O', 'N', 'I', 'N', 'F', 'O', '='};

  ret = memcmp(ptr, leconinfo, sizeof(leconinfo));
  if (ret) {
    dprintf("did not receive leconinfo\n");
    return -1;
  }

  ptr += sizeof(leconinfo);

  int offset = 0;
  uint16_t p_connection_id = (ptr[offset] << 8) | ptr[offset + 1];
  if (p_connection_id != connection_id) {
    dprintf("unexpected connection_id %04x\n", p_connection_id);
  }
  offset += 2;
  offset++; /* comma */
  memcpy(bd_address, &ptr[offset], 6);
  offset += 6;
  /* ignore the rest */

  return 0;
}

#define MAX_NUM_SECBONDDEVS 8 /* max. number for PAN1762 and PAN1760A */

static struct {
  uint8_t bd_address[6];
  uint8_t auth_info;
} secbonddevs_buf[MAX_NUM_SECBONDDEVS];

static int at_secbonddevs(void)
{
  // clang-format off
  uint8_t at_secbonddev[] = {
    'A', 'T', '+', 'S', 'E', 'C', 'B', 'O', 'N', 'D', 'D', 'E', 'V', 'S', ' ', END
  };
  // clang-format on
  uint8_t status_code;
  int i;
  int ret;

  dprintf("retrieve bonding information\n");

  ret = at_transfer_command_status(at_secbonddev, sizeof(at_secbonddev), &status_code);
  if (ret == 0x0a) {
    xprintf("no error - no bonding information available\n");
    return 0;
  }
  if (ret) {
    dprintf("at_transfer_command_status() failed\n");
    return -1;
  }

  int count = status_code;
  for (i = 0; i < count; i++) {
    unsigned int length;

    ret = at_receive_line(buf, sizeof(buf), &length, 1000);
    if (ret) {
      dprintf("at_receive_line() failed\n");
      return -1;
    }

    uint8_t *ptr = buf;

    static uint8_t secbonddevslast[] = {'+', 'S', 'E', 'C', 'B', 'O', 'N', 'D', 'D', 'E', 'V', 'S', 'L', 'A', 'S', 'T', '='};
    ret = memcmp(ptr, secbonddevslast, sizeof(secbonddevslast));
    if (!ret) {
      int offset = 0;
      ptr += sizeof(secbonddevslast);
      memcpy(&secbonddevs_buf[i].bd_address[0], &ptr[offset], 6);
      offset += 6;
      secbonddevs_buf[i].auth_info = ptr[offset++];
      xprintf("SECBONDDEVSLAST: bd_addr " BD_ADDR_FORMAT ", auth_info %02x\n", BD_ADDR_ARG(secbonddevs_buf[i].bd_address), secbonddevs_buf[i].auth_info);
      break;
    }

    static uint8_t secbonddevs[] = {'+', 'S', 'E', 'C', 'B', 'O', 'N', 'D', 'D', 'E', 'V', 'S', '='};
    ret = memcmp(ptr, secbonddevs, sizeof(secbonddevs));
    if (!ret) {
      int offset = 0;
      ptr += sizeof(secbonddevslast);
      memcpy(&secbonddevs_buf[i].bd_address[0], &ptr[offset], 6);
      offset += 6;
      secbonddevs_buf[i].auth_info = ptr[offset++];
      xprintf("SECBONDDEVS: bd_addr " BD_ADDR_FORMAT ", auth_info %02x\n", BD_ADDR_ARG(secbonddevs_buf[i].bd_address), secbonddevs_buf[i].auth_info);
      continue;
    }

    /* fixme unexpected */
    warning("fixme - unexpected event received\n");
  }

  return 0;
}

static int server_spp_rx_data_buffer_characteristic(const uint8_t *data, const int len)
{
  dprintf("SERVER_SPP_RX_DATA: buffer characteristic, length %d\n", len);

  printf("'%.*s'\n", len, data);

  return at_gswrupd(application.connection_id);
}

static int server_spp_rx_data_device_name_characteristic_handler(const uint8_t *data, const int len)
{
  dprintf("SERVER_SPP_RX_DATA: device name characteristic, length %d\n", len);

  memcpy(device_name, data, len);
  device_name_len = len;

  return at_gswrupd(application.connection_id);
}

static int le_conn_completed_event_handler(const uint8_t *buf, const int len)
{
  int ret;

  int offset = 0;
  uint8_t status = buf[offset++];
  offset++; /* comma */
  uint16_t connection_id = (buf[offset] << 8) | buf[offset + 1];
  offset += 2;
  offset++; /* comma */
  uint16_t mtu = (buf[offset] << 8) | buf[offset + 1];

  application.mtu_size = mtu;
  application.is_connected = true;
  application.connection_id = connection_id;

  ret = at_leconinfo(application.connection_id, application.bd_address);
  if (ret) {
    dprintf("at_leconinfo() failed\n");
    return -1;
  }

  xprintf("LE_CONN_COMPLETED_EVENT: connection_id %04x, status %02x, bd_addr " BD_ADDR_FORMAT "\n", connection_id, status, BD_ADDR_ARG(application.bd_address));

  return 0;
}

static int server_write_event_ssp_over_ble_client_characteristic_configuration_descriptor_handler(const uint8_t *value)
{
  uint16_t client_characteristic_configuration_descriptor = value[0] | (value[1] << 8);

  dprintf("%04x\n", client_characteristic_configuration_descriptor);

  application.is_indication_enabled = client_characteristic_configuration_descriptor == 0x02 ? true : false;

  return at_gswrupd(application.connection_id);
}

static int server_read_event(uint8_t service, uint8_t characteristic, uint8_t element)
{
  dprintf("service %d, characteristic %d, element %d\n", service, characteristic, element);

  if (service == 3 && characteristic == 1 && element == 2) {
    uint8_t client_characteristic_configuration_descriptor[2] = {0, 0};
    dprintf("is_indication_enabled %d\n", application.is_indication_enabled);
    if (application.is_indication_enabled) {
      client_characteristic_configuration_descriptor[0] = 0x02;
    }
    return at_gsupdval(3, 1, 2, &client_characteristic_configuration_descriptor[0], sizeof(client_characteristic_configuration_descriptor));
  }

  if (service == 1 && characteristic == 1 && element == 1) {
    return at_gsupdval(1, 1, 1, device_name, device_name_len);
  }

  if (service == 1 && characteristic == 2 && element == 1) {
    return at_gsupdval(1, 2, 1, appearance, sizeof(appearance));
  }

  if (service == 2 && characteristic == 1 && element == 1) {
    return at_gsupdval(2, 1, 1, manufacturer_name_string, sizeof(manufacturer_name_string));
  }

  if (service == 2 && characteristic == 2 && element == 1) {
    return at_gsupdval(2, 2, 1, model_number_string, sizeof(model_number_string));
  }

  if (service == 2 && characteristic == 3 && element == 1) {
    return at_gsupdval(2, 3, 1, system_id, sizeof(system_id));
  }

  /* fall through - unknown combination of service / characteristic / element */
  warning("unhandled service %d, characteristic %d, element %d\n", service, characteristic, element);

  return -1;
}

int application_handle_at_event(const uint8_t *ascii, const int ascii_length, const uint8_t *binary, const int binary_length)
{
  int ret;

  ret = strcmp(ascii, "LE_CONN_COMPLETED_EVENT");
  if (!ret) {
    return le_conn_completed_event_handler(binary, binary_length);
  }

  ret = strcmp(ascii, "LE_CONN_UPDATE_EVENT");
  if (!ret) {
    int offset = 0;
    uint8_t status = binary[offset++];
    offset++; /* comma */
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint16_t conn_intrv = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint16_t conn_lat = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint16_t sup_tout = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;

    xprintf("LE_CONN_UPDATE_EVENT: connection_id %04x, conn_intrv %04x, conn_lat %04x, sup_tout %04x\n", connection_id, conn_intrv, conn_lat, sup_tout);

    return 0;
  }

  ret = strcmp(ascii, "LE_DISCONN_EVENT");
  if (!ret) {
    int offset = 0;
    uint8_t reason = binary[offset++];
    offset++; /* comma */
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;

    flush_rb(&application.in);
    memset(&application, 0, sizeof(application));

    xprintf("LE_DISCONN_EVENT: reason %d, connection_id %04x\n", reason, connection_id);

    ret = at_lesppadv_start();
    if (ret) {
      dprintf("at_lesppadv_start() failed\n");
      return -1;
    }

    return 0;
  }

  ret = strcmp(ascii, "SERVER_MTU_CHANGE_EVENT");
  if (!ret) {
    uint16_t mtu = (binary[3] << 8) | binary[4];
    application.mtu_size = mtu;
    xprintf("SERVER_MTU_CHANGE_EVENT: mtu %d\n", mtu);
    return 0;
  }

  ret = strcmp(ascii, "SERVER_SPP_RX_DATA");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++;                                  /* comma */
    offset++;                                  /* P */
    uint8_t service = binary[offset++];        /* Pi */
    offset++;                                  /* + */
    offset++;                                  /* C */
    uint8_t characteristic = binary[offset++]; /* Ci */
    offset++;                                  /* + */
    offset++;                                  /* E */
    uint8_t element = binary[offset++];        /* Ei */
    offset++;                                  /* comma */

    if (service == 1 && characteristic == 1 && element == 1) {
      /* found write to device information characteristic of generic access service */
      return server_spp_rx_data_device_name_characteristic_handler(&binary[offset], binary_length - offset);
    }

    if (service == 3 && characteristic == 1 && element == 1) {
      /* found write to buffer characteristic */
      return server_spp_rx_data_buffer_characteristic(&binary[offset], binary_length - offset);
    }

    /* fall through: unknown request */
  }

  ret = strcmp(ascii, "SERVER_READ_LONG_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++;                                  /* comma */
    offset++;                                  /* P */
    uint8_t service = binary[offset++];        /* Pi */
    offset++;                                  /* + */
    offset++;                                  /* C */
    uint8_t characteristic = binary[offset++]; /* Ci */
    offset++;                                  /* + */
    offset++;                                  /* E */
    uint8_t element = binary[offset++];        /* Ei */
    offset++;                                  /* comma */
    uint16_t value_offset = (binary[offset] << 8) | binary[offset + 1];

    /* fall through - unknown combination of service / characteristic / element */
    warning("SERVER_READ_LONG_EVENT: unhandled service %d, characteristic %d, element %d, value_offset %d\n", service, characteristic, element, value_offset);

    return at_gsrdupd(service, characteristic, element);
  }

  ret = strcmp(ascii, "SERVER_READ_MULTIPLE_EVENT");
  if (!ret) {
    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    uint8_t service;
    uint8_t characteristic;
    uint8_t element;
    offset += 2;
    offset++; /* comma */
    while (1) {
      offset++;                          /* P */
      service = binary[offset++];        /* Pi */
      offset++;                          /* + */
      offset++;                          /* C */
      characteristic = binary[offset++]; /* Ci */
      offset++;                          /* + */
      offset++;                          /* E */
      element = binary[offset++];        /* Ei */

      dprintf("service %d, characteristic %d, element %d\n", service, characteristic, element);

      ret = server_read_event(service, characteristic, element);
      if (ret) {
        /* fixme */
      }

      uint8_t next = binary[offset++];
      if (next != ',') {
        break;
      }
    }

    return at_gsrdupd(service, characteristic, element);
  }

  ret = strcmp(ascii, "SERVER_READ_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++;                                  /* comma */
    offset++;                                  /* P */
    uint8_t service = binary[offset++];        /* Pi */
    offset++;                                  /* + */
    offset++;                                  /* C */
    uint8_t characteristic = binary[offset++]; /* Ci */
    offset++;                                  /* + */
    offset++;                                  /* E */
    uint8_t element = binary[offset++];        /* Ei */

    dprintf("service %d, characteristic %d, element %d\n", service, characteristic, element);

    ret = server_read_event(service, characteristic, element);
    if (ret) {
      /* fixme: send reject*/
    }
    return at_gsrdupd(service, characteristic, element);
  }

  ret = strcmp(ascii, "SERVER_PREP_WRITE_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++;                                  /* comma */
    offset++;                                  /* P */
    uint8_t service = binary[offset++];        /* Pi */
    offset++;                                  /* + */
    offset++;                                  /* C */
    uint8_t characteristic = binary[offset++]; /* Ci */
    offset++;                                  /* + */
    offset++;                                  /* E */
    uint8_t element = binary[offset++];        /* Ei */
    offset++;                                  /* comma */
    uint16_t write_offset = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;

    if (service == 3 && characteristic == 1 && element == 1) {
      dprintf("SERVER_PREP_WRITE_EVENT: service %d, characteristic %d, element %d, write_offset %04x\n", service, characteristic, element, write_offset);
      return at_gswrupd(connection_id);
    }

    /* fall through - unknown combination of service / characteristic / element */
    warning("SERVER_PREP_WRITE_EVENT: unhandled service %d, characteristic %d, element %d\n", service, characteristic, element);
  }

  ret = strcmp(ascii, "SERVER_WRITE_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++;                                  /* comma */
    offset++;                                  /* P */
    uint8_t service = binary[offset++];        /* Pi */
    offset++;                                  /* + */
    offset++;                                  /* C */
    uint8_t characteristic = binary[offset++]; /* Ci */
    offset++;                                  /* + */
    offset++;                                  /* E */
    uint8_t element = binary[offset++];        /* Ei */
    offset++;                                  /* comma */
    const uint8_t *value = &binary[offset];

    if (service == 3 && characteristic == 1 && element == 2) {
      return server_write_event_ssp_over_ble_client_characteristic_configuration_descriptor_handler(value);
    }

    /* fall through - unknown combination of service / characteristic / element */
    warning("SERVER_WRITE_EVENT: unhandled service %d, characteristic %d, element %d\n", service, characteristic, element);
  }

  ret = strcmp(ascii, "SERVER_INDICATION_ACK_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint8_t status = binary[offset++];
    offset++;                                  /* comma */
    offset++;                                  /* P */
    uint8_t service = binary[offset++];        /* Pi */
    offset++;                                  /* + */
    offset++;                                  /* C */
    uint8_t characteristic = binary[offset++]; /* Cj */

    if (connection_id != application.connection_id) {
      dprintf("unexpected connection_id %04x\n", connection_id);
      return -1;
    }
    if (status) {
      dprintf("unexpected status %02x\n", status);
      return -1;
    }
    if (service != 3) {
      dprintf("unexpected service %02x\n", service);
      return -1;
    }
    if (characteristic != 1) {
      dprintf("unexpected characteristic %02x\n", characteristic);
      return -1;
    }

    application.is_sending = false;

    dprintf("SERVER_INDICATION_ACK_EVENT, connection_id %04x\n", connection_id);

    return 0;
  }

  ret = strcmp(ascii, "SEC_BONDING_REQ_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;

    xprintf("SEC_BONDING_REQ_EVENT, connection_id %04x\n", connection_id);

    return at_secbondaccept(connection_id);
  }

  ret = strcmp(ascii, "SEC_BONDING_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint8_t status = binary[offset++];
    offset++; /* comma */
    uint8_t bonding_mode = binary[offset++];
    offset++; /* comma */
    uint8_t auth_info = binary[offset++];

    xprintf("SEC_BONDING_EVENT: connection_id %04x, status %02x, bonding_mode %02x, auth_info %02x\n", connection_id, status, bonding_mode, auth_info);

    return 0;
  }

  ret = strcmp(ascii, "SEC_BONDING_FAILED_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint8_t reason = binary[offset++];

    xprintf("SEC_BONDING_FAILED_EVENT, connection_id %04x, reason %02x\n", connection_id, reason);

    return 0;
  }

  ret = strcmp(ascii, "SEC_ENCRYPTION_CHANGE_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset += 2;
    offset++; /* comma */
    uint8_t status = binary[offset++];

    xprintf("SEC_ENCRYPTION_CHANGE_EVENT: status %02x\n", status);

    ret = at_leresolvaddr(application.bd_address);
    if (ret) {
      dprintf("at_leresolvaddr() failed\n");
      /* fall through */
    }

    return 0;
  }

  ret = strcmp(ascii, "SEC_IDENTITY_ADDRESS_RECVD_EVENT");
  if (!ret) {

    int offset = 0;
    uint8_t address_type = binary[offset++];
    uint8_t bd_address[6];
    memcpy(bd_address, &binary[offset], 6);
    offset += 6;

    xprintf("SEC_IDENTITY_ADDRESS_RECVD_EVENT: address_type %02x, bd_addr " BD_ADDR_FORMAT "\n", address_type, BD_ADDR_ARG(bd_address));

    return 0;
  }

  ret = strcmp(ascii, "SECDISPKEYEV");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];

    xprintf("SECDISPKEYEV: connection_id %04x\n", connection_id);

    return at_seccfmpkey(connection_id);
  }

  ret = strcmp(ascii, "LE_DATA_LENGTH_CHANGE_EVENT");
  if (!ret) {

    int offset = 0;
    uint8_t status = binary[offset++];
    offset++; /* comma */
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset++; /* comma */
    uint16_t tx_octets = (binary[offset] << 8) | binary[offset + 1];
    offset++; /* comma */
    uint16_t tx_time = (binary[offset] << 8) | binary[offset + 1];
    offset++; /* comma */
    uint16_t rx_octets = (binary[offset] << 8) | binary[offset + 1];
    offset++; /* comma */
    uint16_t rx_time = (binary[offset] << 8) | binary[offset + 1];

    dprintf("LE_DATA_LENGTH_CHANGE_EVENT: connection_id %04x, status %02x, tx_octets %d, tx_time %d, rx_octets %d, rx_time %d\n", connection_id, status, tx_octets, tx_time, rx_octets, rx_time);

    return 0;
  }

  ret = strcmp(ascii, "LE_EXTENDED_CHANNEL_SELECTION_ALGO_EVENT");
  if (!ret) {

    int offset = 0;
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset++; /* comma */
    uint8_t algorithm = binary[offset++];

    xprintf("LE_EXTENDED_CHANNEL_SELECTION_ALGO_EVENT: connection_id %04x, algorithm %02x\n", connection_id, algorithm);

    return 0;
  }

  ret = strcmp(ascii, "LE_PHY_UPDATE_EVENT");
  if (!ret) {

    int offset = 0;
    uint8_t status = binary[offset++];
    offset++; /* comma */
    uint16_t connection_id = (binary[offset] << 8) | binary[offset + 1];
    offset++; /* comma */
    uint8_t algorithm = binary[offset++];
    offset++; /* comma */
    uint8_t tx_phy = binary[offset++];
    offset++; /* comma */
    uint8_t rx_phy = binary[offset++];

    xprintf("LE_PHY_UPDATE_EVENT: connection_id %04x, status %02x, tx_phy %02x, rx_phy %02x\n", connection_id, status, tx_phy, rx_phy);

    return 0;
  }

  warning("unhandled event '%s'\n", ascii);
  return 0;
}

int application_init(const char *new_device_name)
{
  uint8_t authentication_requirement = 0x00;
  uint16_t mtu = 160;
  int ret;

  /* no safety check here */
  if (new_device_name) {
    device_name_len = strlen(new_device_name);
    memcpy(device_name, new_device_name, device_name_len);
  }

  dprintf("\n");

  application.mtu_size = mtu;

  ret = at_lesppinit(authentication_requirement, &appearance[0], mtu);
  if (ret) {
    dprintf("at_lesppinit() failed\n");
    return 0;
  }

#if 0
  ret = at_secsetparam();
  if (ret) {
    dprintf("at_secsetparam() failed\n");
    return 0;
  }

  ret = at_secbonddevs();
  if (ret) {
    dprintf("at_secbonddevs() failed\n");
    return -1;
  }
#endif

  ret = at_lesppadv_start();
  if (ret) {
    dprintf("at_lesppadv_start() failed\n");
    return 0;
  }
  
  return 0;
}

void application_shutdown(void)
{
  int ret;

  if (application.is_connected == false) {
    return;
  }

  ret = at_send_lediscon(application.bd_address);
  if (ret) {
    dprintf("at_send_lediscon() failed\n");
    return;
  }
}

int application_handle_keyboard_input(int key)
{
  if (!application.is_connected) {
    dprintf("not connected - doing nothing\n");
    return 0;
  }

  if (!application.is_indication_enabled) {
    dprintf("indication not enabled - doing nothing\n");
    return 0;
  }

  add_to_rb(&application.in, (uint8_t)key);

  return 0;
}

/* application handling again */
int application_process(void)
{
  int len = 0;
  int ret;

  if (!application.is_connected) {
    //    dprintf("not connected - doing nothing\n");
    return 0;
  }

  if (!application.is_indication_enabled) {
    //    dprintf("indication not enabled - doing nothing\n");
    return 0;
  }

  if (application.is_sending == true) {
    //    dprintf("currently sending - doing nothing\n");
    return 0;
  }

  while (len < application.mtu_size) {
    ret = remove_from_rb(&application.in, &application.buf[len]);
    if (ret) {
      break;
    }
    len++;
  }

  if (!len) {
    // dprintf("no data to send - doing nothing\n");
    return 0;
  }

  if (application.is_sending == true) {
    dprintf("already sending\n");
    return -1;
  }

  uint16_t timeout = 1000;

  ret = at_gssrvind(3, 1, timeout, application.buf, len);
  if (ret) {
    dprintf("at_gssrvind() failed\n");
    return -1;
  }

  application.is_sending = true;

  //  dprintf("length %d\n", length);

  return 0;
}
