/****************************************************************************
*   COPYRIGHT 2019 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

#include "application.h"
#include "getopt.h"
#include "board_specific.h" /* for NVM_SIZE ... */
#include "tcu_api.h"
#include "application_flash_partitions.h"

struct nvm_partition ble_partitions[] = {
  { nvm_reserved, 4096 },            /* A-side header */
  { nvm_application_a0, 7 * 4096 },  /* A0 application*/
  { nvm_application_a1, 12 * 4096 }, /* A1 application*/
  { nvm_reserved, 4096 },            /* B-side header */
  { nvm_application_b0, 7 * 4096 },  /* B0 application*/
  { nvm_application_b1, 12 * 4096 }, /* B1 application*/
  { nvm_key_store_data_1, 4096 },    /* Key store data 1 */
  { nvm_key_store_data_2, 4096 },    /* Key store data 2 */
  { nvm_partition_end, },            /* end marker - don't remove */
};

extern struct nvm_partition other_partitions[];
extern struct nvm_partition system_partitions[];

struct nvm_partition *const nvm_partitions[] = {
    &system_partitions[0], &other_partitions[0], &ble_partitions[0], NULL, /* end marker - don't remove */
};

static uint8_t tcu_implementation_buffer[MAX_BUFFER_SIZE_BLE];
uint8_t *uart_buffer = &tcu_implementation_buffer[0];
uint32_t uart_buffer_len = sizeof(tcu_implementation_buffer);
uint8_t *tcu_buffer = &tcu_implementation_buffer[0];
uint32_t tcu_buffer_len = sizeof(tcu_implementation_buffer);

uint8_t init_done = 0;

static HANDLE hRxUart;

#include "hci_api.h"

static bool erase_flash = false;

static int factory_reset(uint8_t *bd_addr)
{
  int ret;

  ret = hci_api_erase_storage_device();
  if (ret) {
    dprintf("hci_api_erase_storage_device() failed\n");
    goto error;
  }

  ret = hci_api_write_data_to_storage_device(0x02, 0x06, bd_addr);
  if (ret) {
    dprintf("hci_api_write_data_to_storage_device() failed\n");
    goto error;
  }

  ret = hci_api_close_storage_flash();
  if (ret) {
    dprintf("hci_api_close_storage_flash() failed\n");
    goto error;
  }

  xprintf("%s(): factory reset ok, Bluetooth device address is " BD_ADDR_FORMAT_STRING "\n", __func__, BD_ADDR_ARGUMENTS_LE(bd_addr));
  exit(0);

error:
  xprintf("%s(): factory reset failed\n", __func__);
  exit(0);
}

static int hci_to_tcu_init(enum tsb_bluetooth_chip_version *version)
{
  int ret;

  uint8_t bd_addr[TCU_SIZE_BDADDR];
  uint8_t buf[TCU_SIZE_BDADDR];
  int i;

  /* This is not strictly necessary but will make sure that the device is in a clean state */
  ret = hci_api_reset();
  if (ret) {
    dprintf("hci_api_reset() failed\n");
    return -1;
  }

  /* give TC3566x time to come up again */
  Sleep(100);

  ret = hci_api_read_firmware_version(version);
  if (ret) {
    dprintf("hci_api_read_firmware_version() failed\n");
    return -1;
  }

  if (*version != TC35680_1_002_ROM) {
    xprintf("%s(): This is not a PAN1762 device\n", __func__);
    exit(0);
  }

  /* Beware: At this point RTS pin must be low. After the PAN module has received this command,
  it will honour the RTS line and if it is high, it will not send the confirmation for
  this command any more.
  In our case the FTDI adapter will configure the RTS to low in ftdi_init().
  For microcontrollers this same has to be done.
  If no RTS/CTS flow control is configured, the corresponding pin may be left unconfigured
  and depending on the default pin state, it might be even configured as GPIO - input - pullup.
  This will prevent the PAN module from sending the confirmation of this command.
  So make sure that the RTS line is low before calling this command. */
  ret = hci_api_set_rtscts();
  if (ret) {
    dprintf("hci_api_set_rtscts() failed\n");
    return -1;
  }

  ret = ftdi_enable_flow_control();
  if (ret) {
    dprintf("ftdi_enable_flow_control() failed\n");
    return -1;
  }

  ret = hci_api_set_baudrate(*version, HCI_API_BAUDRATE_921600);
  if (ret) {
    dprintf("hci_api_set_baudrate() failed\n");
    return -1;
  }

  /* reconfigure baud rate to desired rate */
  ret = ftdi_set_baudrate(921600);
  if (ret) {
    dprintf("ftdi_set_baudrate() failed\n");
    return -1;
  }

  /* give TC3566x time to change the baudrate */
  Sleep(25);

  ret = hci_api_open_storage_flash(0x00000001);
  if (ret) {
    dprintf("hci_api_open_storage_flash() failed\n");
    return -1;
  }

  ret = hci_api_read_data_from_storage_device(0x02, 0x06, bd_addr);
  if (ret) {
    dprintf("hci_api_read_data_from_storage_device() failed\n");
    return -1;
  }

  static uint8_t all_00[] = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00};
  static uint8_t all_ff[] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

  ret = memcmp(all_00, bd_addr, 6);
  if (ret) {
    ret = memcmp(all_ff, bd_addr, 6);
  }
  if (!ret) {
    xprintf("%s(): Bluetooth device address not set correctly. Please use EasyStandalone to set a correct Bluetooth device address\n",
            __func__);
    exit(0);
  }

  if (erase_flash == true) {
    return factory_reset(bd_addr);
  }

  xprintf("%s(): Bluetooth device address " BD_ADDR_FORMAT_STRING "\n", __func__, BD_ADDR_ARGUMENTS_LE(bd_addr));

  /* byte order in eeprom is reversed to what HCI controller expects */
  for (i = 0; i < 6; i++) {
    buf[5 - i] = bd_addr[i];
  }

  ret = hci_api_write_bd_addr(&buf[0]);
  if (ret) {
    dprintf("hci_api_write_bd_addr() failed\n");
    return -1;
  }

  ret = hci_api_switch_to_tcu_mode(*version);
  if (ret) {
    dprintf("hci_api_switch_to_tcu_mode() failed\n");
    return -1;
  }

  /* a minimal delay is needed to allow TC356xx to switch to TCU mode */
  Sleep(30);

  return 0;
}

#include "ble_api_message.h"

static void shutdown_application(void)
{
  enum ble_api_result result;

  if (application.is_connected == false) {
    return;
  }

  /* if a connection exists, gracefully disconnect it before terminating the
   * application */

  result = ble_api_disconnect(&application.connection.bd_addr[0]);
  if (result != LE_API_SUCCESS) {
    dprintf("ble_api_disconnect() failed\n");
    /* fall through */
  }
}

static char *device_name = "PAN1762-USB";
static int device_name_len;

/* application handling */
int main(int argc, char *argv[])
{
  enum tsb_bluetooth_chip_version version;
  enum ble_api_result result;
  int opt;
  bool erase_key_storage = false;
  int ret;

  while ((opt = getopt(argc, argv, "sfn:")) != -1) {
    switch (opt) {
      case 's':
        erase_key_storage = true;
        break;
      case 'f':
        erase_flash = true;
        break;
      case 'n':
        device_name = optarg;
        break;
      default:
        exit(1);
    }
  }

  device_name_len = (int)strlen(device_name);

  ret = ftdi_init(&hRxUart);
  if (ret) {
    dprintf("ftdi_init() failed\n");
    return -1;
  }

  ret = hci_to_tcu_init(&version);
  if (ret) {
    dprintf("hci_to_tcu_init() failed\n");
    return -1;
  }

  if (version == TC35680_1_002_ROM) {
    result = ble_api_configure_bt_chip_mode(version, LEGACY);
    if (result != LE_API_SUCCESS) {
      xprintf("ble_api_configure_bt_chip_mode() failed\n");
      return -1;
    }
  }

  result = ble_api_init(version);
  if (result != LE_API_SUCCESS) {
    xprintf("ble_api_init() failed\n");
    return -1;
  }

  ret = nvm_init(NVM_SIZE);
  if (ret) {
    xprintf("nvm_init() failed\n");
    return -1;
  }

  ret = key_store_ble_init();
  if (ret) {
    dprintf("key_store_ble_init() failed\n");
    return -1;
  }

  if (erase_key_storage == true) {
    ret = key_store_ble_reset();
    if (ret) {
      xprintf("key_store_ble_reset() failed\n");
      return -1;
    }
  }

  ret = application_init(device_name, device_name_len);
  if (ret) {
    xprintf("application_init() failed\n");
    return -1;
  }

  HANDLE hStdin;

  hStdin = GetStdHandle(STD_INPUT_HANDLE);
  if (hStdin == INVALID_HANDLE_VALUE) {
    xprintf("GetStdHandle() failed\n");
    return -1;
  }

  DWORD fdwMode;

  ret = GetConsoleMode(hStdin, &fdwMode);
  if (!ret) {
    xprintf("GetConsoleMode() failed\n");
    return -1;
  }

  fdwMode = fdwMode & ~(ENABLE_LINE_INPUT | ENABLE_ECHO_INPUT | ENABLE_MOUSE_INPUT | ENABLE_PROCESSED_INPUT);

  ret = SetConsoleMode(hStdin, fdwMode);
  if (!ret) {
    xprintf("SetConsoleMode() failed\n");
    return -1;
  }

  FlushConsoleInputBuffer(hStdin);
  if (!ret) {
    xprintf("FlushConsoleInputBuffer() failed\n");
    return -1;
  }

  bool shift_pressed = false;
  bool ctrl_pressed = false;

  while (1) {

    /* check if application needs to do any Bluetooth-related processing */
    ret = application_process();
    if (ret) {
      xprintf("application_process() failed\n");
      return -1;
    }

    HANDLE handles[] = {
        hRxUart, hStdin,
    };
    DWORD nCount = sizeof(handles) / sizeof(handles[0]);
    ret = WaitForMultipleObjects(nCount, &handles[0], FALSE, INFINITE);

    switch (ret) {
      case WAIT_TIMEOUT: {
        xprintf("WAIT_TIMEOUT @ WaitForMultipleObjects()\n");
        continue;
      }
      case WAIT_FAILED: {
        xprintf("WAIT_FAILED @ WaitForMultipleObjects()\n");
        goto out;
      }
      case WAIT_OBJECT_0: {
        enum ble_api_result result;
        result = ble_api_message_handler(server, client);
        if (result) {
          xprintf("ble_api_message_handler() failed\n");
          goto out;
        }
        break;
      }
      case WAIT_OBJECT_0 + 1: {
        INPUT_RECORD record;
        DWORD numRead;

        if (!ReadConsoleInput(hStdin, &record, 1, &numRead)) {
          xprintf("ReadConsoleInput() failed\n");
          goto out;
        }

        if (record.EventType != KEY_EVENT) {
          // don't care about other console events
          continue;
        }

        if (record.Event.KeyEvent.wVirtualKeyCode == VK_SHIFT) {
          if (record.Event.KeyEvent.bKeyDown) {
            shift_pressed = true;
          } else {
            shift_pressed = false;
          }
          continue;
        }

        if (record.Event.KeyEvent.wVirtualKeyCode == VK_CONTROL) {
          if (record.Event.KeyEvent.bKeyDown) {
            ctrl_pressed = true;
          } else {
            ctrl_pressed = false;
          }
          continue;
        }

        if (!record.Event.KeyEvent.bKeyDown) {
          // really only care about keydown
          continue;
        }

        if (ctrl_pressed == true && record.Event.KeyEvent.wVirtualKeyCode == 0x43) {
          shutdown_application();
          ExitProcess(0);
          continue;
        }

        int key = record.Event.KeyEvent.uChar.AsciiChar;

        if (shift_pressed) {
          key = toupper(key);
        }

        ret = application_handle_keyboard_input(key);
        if (ret) {
          dprintf("application_handle_keyboard_inputs() failed\n");
        }
      }
    }
  }

out:
  printf("Application ended unexpectedly - did you push the reset button?\n");
  return 0;
}

#include "tcu_api.h"
int tcu_application_store_unexpected_event_callback(uint8_t *buf, uint32_t length)
{
  dprintf("\n");
  __debugbreak();
  return 0;
}
