/****************************************************************************
*   COPYRIGHT 2019 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

#ifndef __APPLICATION_H__
#define __APPLICATION_H__

/**
* @file configurations\implementation\application\include\application.h
* @brief Header file for the common application parts shared between the peripheral and central demo
* @details This header files contains the common application parts that are
* shared between the peripheral and central demo. Please note that this code
* is also shared between the host mode and standalone mode demos, so no
* platform-specific code is allowed here. It has to be put into "environment.h".
*
**/

#include "types.h"
#include "environment.h"
#include "ble_api.h"
#include "debug.h"
/* This comes from deployment environment and should probably be named shared.h instead... Must be last,
otherwise the MSVC include search order preferes user_configuration.h from the host mode deployment. */
#include "internal_apis.h"

/*! Definition of the maximum size of a ringbuffer instance */
#define RING_SIZE 512
/*! Typedef for a position in the ringbuffer. Should be the "native" type of the processor */
typedef uint32_t ring_pos_t;

/*! Structure for a ringbuffer */
struct ringbuffer {
  volatile ring_pos_t ring_head;         /*!< "Pointer" to the head of the rinbuffer */
  volatile ring_pos_t ring_tail;         /*!< "Pointer" to the tail of the ringbuffer */
  volatile uint8_t ring_data[RING_SIZE]; /*!< Data buffer for the ringbuffer */
};

/**
* @brief Function that flushes a ringbuffer
*
* @param[in] rb Pointer to the ringbuffer
*
* @return 0 if success, -1 if error
*/
void flush_rb(struct ringbuffer *rb);

/**
* @brief Function that adds one element to a ringbuffer
*
* @param[in] rb Pointer to the ringbuffer
* @param[in] c Element to be added
*
* @return 0 if success, -1 if error
*/
int add_to_rb(struct ringbuffer *rb, uint8_t c);

/**
* @brief Function that removes one element from a ringbuffer
*
* @param[in] rb Pointer to the ringbuffer
* @param[in] c Pointer where the removed element is stored
*
* @return 0 if success, -1 if error or ringbuffer is empty
*/
int remove_from_rb(struct ringbuffer *rb, uint8_t *c);

/**
* @brief Function that peeks at the first element of a ringbuffer
*
* @param[in] rb Pointer to the ringbuffer
* @param[in] c Pointer where the first element is stored
*
* @return 0 if success, -1 if error or ringbuffer is empty
*/
int peek_into_rb(struct ringbuffer *rb, uint8_t *c);

#ifdef SPEED_TEST
/*! Structure that contains variables specific to the speed test */
struct speed_test {
  bool is_enabled; /*!< Indicates if the speed test is enabled */
  int total_bytes; /*!< Counter for the total number of bytes that have been processed */
};
#endif

/*! Structure that contains the application specific data */
struct application {
  uint8_t buf[SER_ATT_MTU_SIZE_MAX];    /*!< Intermediate buffer that is used for data sending and reception */
  bool is_connected;                    /*!< Indicates if a connection is currently established */
  struct ble_connection_parameters connection; /*!< Details for the established connection */
  int mtu_size;                         /*!< MTU size of the current connection */
  bool is_spp_connected;                /*!< Indicates if a "virtual" SPPoverBLE connection has been established */
  bool is_sending;                      /*!< Indicates if the device is currently sending */
                                        //  bool encryption_setup_failed;         /*!< Indicates if the encryption setup has failed  */
                                        //  bool is_encrypted;                    /*!< Indicates if the connection is encrypted */
  struct ringbuffer in;                 /*!< Ringbuffer for the intermediate storage of keypresses */
#ifdef SPEED_TEST
  bool is_configuration_enabled;      /*!< Indicates if the application configuration (peripheral/central) has been enabled */
  struct timer_data timer_speed_test; /*!< A timer to be used for calculating and printing the statistics */
  struct speed_test tx;               /*!< Data storage for the Tx speed test */
  struct speed_test rx;               /*!< Data storage for the Tx speed test */
#endif
};

/*! Declaration of the global application data structure */
extern struct application application;

/**
* @brief Function that initialises the application.
*
* @param[in] device_name Pointer to a device name to be used
* @param[in] device_name_len Length of the device name
*
* @return 0 if success, -1 if error
*/
int application_init(const uint8_t *device_name, int device_name_len);

/**
* @brief Function that handles regular application processing. It is called whenever the
* application has been woken up, either due to Bluetooth handling, UART handling or application
* handling. For the peripheral demo, keypresses to be sent to the remote device, for example, are collected in a
* ringbuffer. When this function is called, the application should check if data can be send and if so, initiate the
* send operation.
*
* @return 0 if success, -1 if error
*/
int application_process(void);

#ifdef SPEED_TEST
int configuration_init(const uint8_t *device_name, int device_name_len);
int configuration_start(void);
int configuration_stop(void);
int configuration_speed_test_tx(uint8_t *data, uint16_t length, uint16_t connection_handle);
int configuration_process(void);
#endif

/**
* @brief Function that handles any keyboard input. For all demos, keypresses are collected
* in a ringbuffer for later processing in application_process().
*
* @param[in] key ASCII value of the key that was pressed
*
* @return 0 if success, -1 if error
*/
int application_handle_keyboard_input(int key);

/*! Declaration of global server profile, used by ble_api_message_handler() */
extern struct ble_api_srv_profile *server;
/*! Declaration of global client profile, used by ble_api_message_handler() */
extern struct ble_api_cli_profile *client;

/*! Definition of a format string for printing a Bluetooth device address */
#define BD_ADDR_FORMAT_STRING "%02x:%02x:%02x:%02x:%02x:%02x"

/*! Definition of an argument list for printing a Bluetooth device address in little-endian format */
#define BD_ADDR_ARGUMENTS_BE(x) x[5], x[4], x[3], x[2], x[1], x[0]

/*! Definition of an argument list for printing a Bluetooth device address in big-endian format */
#define BD_ADDR_ARGUMENTS_LE(x) x[0], x[1], x[2], x[3], x[4], x[5]

#endif