/****************************************************************************
*   COPYRIGHT 2019 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

/**
* @file debug.h
*
* @brief Contains definitions and function declarations for debug handling
*/
#ifndef __DEBUG_H__
#define __DEBUG_H__

#include "types.h"
#include "version_specific.h"

/*! assert() is currently a no-op */
#define assert(...)

/**
* @brief Initialize the debugging subsystem.
*
* @return 0 if success, -1 if error
*/
int debug_init(void);

/**
* @brief Send a message to the debugging subsystem.
*
* This functions blocks until the message was sent.
*
* @param[in] format Format of the message, similar to printf()
*
* @return 0 if success, -1 if error
*/
int debug_send(const char *format, ...);

#ifndef DEBUG
/*! If DEBUG is not defined, vPrintf() is a no-op */
#define vPrintf(...)
#else
/*! If DEBUG is defined, use debug functions to output information */
#define vPrintf(...)                                                                                                                  \
  do {                                                                                                                                \
    debug_send("%s(): ", __func__);                                                                                                   \
    debug_send(__VA_ARGS__);                                                                                                          \
  } while (0)

#endif

#ifndef NDEBUG
/*! If DEBUG is defined, use debug functions to output information */
#define warning(...)                                                                                                                  \
  do {                                                                                                                                \
    debug_send(__VA_ARGS__);                                                                                                          \
  } while (0)
#else
/*! If DEBUG is not defined, warning() is a no-op */
#define warning(...)
#endif

#ifndef NDEBUG
/*! If DEBUG is defined, use debug functions to output information */
#define bs_warning(...)                                                                                                               \
  do {                                                                                                                                \
    debug_send("[W] %s(): ", __func__);                                                                                               \
    debug_send(__VA_ARGS__);                                                                                                          \
  } while (0)
#else
/*! If DEBUG is not defined, bs_warning() is a no-op */
#define bs_warning(...)
#endif

/*! Definition of a printf macro that will output directly to UART1, without further processing. */
#define xprintf(...)                                                                                                                  \
  do {                                                                                                                                \
    debug_send(__VA_ARGS__);                                                                                                          \
  } while (0)

#ifndef DEBUG
/*! If DEBUG is not defined, dprintf() is a no-op */
#define dprintf(...)
#else
/*! If DEBUG is defined, use debug functions to output information */
#define dprintf(...)                                                                                                                  \
  do {                                                                                                                                \
    debug_send("%s(): ", __func__);                                                                                                   \
    debug_send(__VA_ARGS__);                                                                                                          \
  } while (0)
#endif

/*! empty default debug macro definition a no-op */
#define gatt_bs_printf(...)
#define gap_bs_printf(...)
#define smp_bs_printf(...)
#define spp_bs_printf(...)
#define ven_bs_printf(...)
#define hci_bs_printf(...)
#define buff_bs_printf(...)
#define tcu_bs_printf(...)

#ifndef BD_ADDR_REV_ARG
#define BD_ADDR_REV_ARG(x) x[5], x[4], x[3], x[2], x[1], x[0]
#endif

#ifndef IRK_FORMAT
#define IRK_FORMAT "%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x%x"
#endif

#ifndef IRK_ARG
#define IRK_ARG(x) x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], x[8], x[9], x[10], x[11], x[12], x[13], x[14], x[15]
#endif

#ifndef OOB_REV_ARG
#define OOB_REV_ARG(x) x[15], x[14], x[13], x[12], x[11], x[10], x[9], x[8], x[7], x[6], x[5], x[4], x[3], x[2], x[1], x[0]
#endif

#endif
