/****************************************************************************
*   COPYRIGHT 2019 All rights reserved
*   Panasonic Industrial Devices Europe GmbH
****************************************************************************/

//#define DEBUG

/* fixme: add data length extension */

#include "application.h"
#include "ble_api.h"
#include "system.h"
#include "osal.h"
#include "host_api.h"
#include "host_api_defs.h"
#include "host_api_add.h"
#include "board.h"
#include "Driver_LowPower_TC35680.h"
#include "uart1_function.h"
#include "debug.h"

/* This does not belong here */
uint8_t init_done = 0;

static uint8_t device_name[] = "PAN1762-USB";
static int device_name_len = sizeof(device_name) - 1;

/* the following is TC35680 specific and probably should not be here */
#include "internal_apis.h"

struct nvm_partition system_partitions[] = {
    {nvm_reserved, 2},
    {nvm_bd_address, 6},
    {nvm_reserved, 4088},
    {
        nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition other_partitions[] = {
    {nvm_key_store_data, 4096}, /* Key store data */
    {nvm_user_area, 2 * 4096},  /* User Area */
    {
        nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition ble_partitions[] = {
    {nvm_reserved, 4096},          /* A-side header */
    {nvm_reserved, 4096},          /* B-side header */
    {nvm_reserved, 2 * 13 * 4096}, /* Application */
    {
        nvm_partition_end,
    }, /* end marker - don't remove */
};

struct nvm_partition *const nvm_partitions[] = {
    &system_partitions[0], &other_partitions[0], &ble_partitions[0], NULL, /* end marker - don't remove */
};

/* This function reads the pre-programmed Bluetooth device address from the
built-in flash. Surprisingly, this also initializes the Bluetooth core with
this address. So this procedure is mandatory to be executed. */
static int init_bd_address(void)
{
#define BD_ADDR_LEN 0x06
#define BD_ADDR_OFFSET 0x02
  uint8_t bd_address[BD_ADDR_LEN];
  int ret;

  ret = SYS_API_QSPIStorageOpen(0x01, 0x00);
  if (ret) {
    dprintf("SYS_API_QSPIStorageOpen() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageReadSpecifiedAddr(BD_ADDR_LEN, BD_ADDR_OFFSET);
  if (ret) {
    dprintf("SYS_API_QSPIStorageReadSpecifiedAddr() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageRead(BD_ADDR_OFFSET, BD_ADDR_LEN, &bd_address[0]);
  if (ret) {
    dprintf("SYS_API_QSPIStorageRead() failed\n");
    return -1;
  }

  ret = SYS_API_QSPIStorageClose();
  if (ret) {
    dprintf("SYS_API_QSPIStorageClose() failed\n");
    return -1;
  }

  dprintf("\n");
  return 0;
}

/* This is the main function. It does the basic initialization and then goes
into the main message processing loop. The "system" and "message" handling
functions come straight from the Bluetooth SDK. */
C2_ENTRY int C2_Task(void)
{
  enum bt_tsb_bluetooth_chip_version version;
  enum ble_api_result result;
  unsigned int flag = 0;
  uint16_t pmu_cause;
  uint16_t crw_rst_cause;
  int ret;

  DATA_INIT();

  uart1_init();

  ret = SYS_API_GetWakeUpCauseValue(NULL, &pmu_cause, &crw_rst_cause);
  if (ret) {
    dprintf("SYS_API_GetWakeUpCauseValue() failed\n");
    goto fatal_error;
  }
  
  if (pmu_cause & 0x01) {
    dprintf("wake up from deep sleep\n");
  } else if (pmu_cause & (0x01 << 3)) {
    dprintf("wake up from cold boot\n");
  } else {
    dprintf("unknown wake up reason\n");
  }

  ret = init_bd_address();
  if (ret) {
    dprintf("init_bd_address() failed\n");
    goto fatal_error;
  }

  ret = system_read_version(&version);
  if (ret) {
    dprintf("system_read_version() failed\n");
    goto fatal_error;
  }

  C2_Patch();

  /* do not set up low power mode - UART 1 does not wake up system */

  result = ble_api_configure_bt_chip_mode(version, LEGACY);
  if (result) {
    vPrintf("Mode Init failed\n");
    goto fatal_error;
  }

  result = ble_api_init(version);
  if (result) {
    dprintf("ble_api_init() failed\n");
    goto fatal_error;
  }

  init_done |= BLE_INIT;

  ret = nvm_init(NVM_SIZE);
  if (ret) {
    dprintf("nvm_init() failed\n");
    goto fatal_error;
  }

  ret = key_store_ble_init();
  if (ret) {
    dprintf("key_store_ble_init() failed\n");
    goto fatal_error;
  }

  /* The peripheral and central demo need different init functions */
  ret = application_init(device_name, device_name_len);
  if (ret) {
    dprintf("application_init() failed\n");
    goto fatal_error;
  }

  while (1) {

    application_process();

    event_wait(&flag);

    /* we do not handle any low power events here */
    /* we do not handle any rtc events here either */

    if ((flag & C2_MSG_INT_BIT)) {
      /* ble_api_message_handler is called if received event is BLE event */
      ret = ble_api_message_handler(server, client);
      if (ret) {
        dprintf("ble_api_message_handler() failed\n");
        /* fall through */
      }
    }
  }

fatal_error:
  warning("fatal error - application did not start\n");
  while (1) {
    /* fixme */
  }
}

int uart1_receive_bytes_callback_isr(uint8_t *ptr, int len)
{
  int i;
  int ret;

  for (i = 0; i < len; i++) {
    ret = application_handle_keyboard_input(ptr[i]);
    if (ret) {
      return -1;
    }
  }

  return 0;
}

#include "mini-printf.h"

static uint8_t buf[128];
int debug_send(const char *format, ...)
{
  va_list args;
  int ret;

  va_start(args, format);
  ret = mini_vsnprintf((char *)buf, sizeof(buf), format, args);
  va_end(args);

  uart1_send(buf, ret);
  return 0;
}

void data_receive_debug_callback(uint8_t *buf, uint32_t length)
{
#if 0
  uint32_t i;

  for (i = 0; i < length; i++) {
    /* This will print raw data in ASCII format.
     for binary data can use %c format specifier */
    debug_send("0x%x ", buf[i]);
  }

  debug_send("\n");
#endif
}

void tcu_message_debug_output(const char *prefix, const uint8_t *buf, uint32_t length)
{
#if 0
  static char debug_buf[256];
  
  int i;
  char *ptr = &debug_buf[0];
  int len = sizeof(debug_buf);
  int ret;
  
  ret = mini_snprintf(ptr, len, "%03d %s ", length, prefix);
  ptr += ret; len -= ret;
  
  for (i = 0; i < length; i++) {
    ret = mini_snprintf(ptr, len, "%02x ", buf[i]);
    ptr += ret;  len -= ret;
    if (!len) {
      break;
    }
  }

  uart1_printf("%s\n", debug_buf);
#endif
}

void ble_application_default_event_callback(struct bt_cmn_msg *bt_cmn_msg, uint8_t *data)
{
  warning("\n");
}

void ble_application_uart_message_callback(struct bt_cmn_msg *bt_cmn_msg, uint8_t *data)
{
  application_process();
}

extern int message_timer_callback(void *timer);

void ble_application_timer_message_callback(struct bt_cmn_msg *bt_cmn_msg, uint8_t *data)
{
  message_timer_callback((void *)bt_cmn_msg->user_data);
}

void ble_application_event_handling_failed_callback(uint8_t service_id, uint8_t opcode, uint8_t error)
{
  dprintf("\n");
  __debugbreak();
}