/*
* COPYRIGHT (C) 2016
* TOSHIBA CORPORATION STORAGE & ELECTRONIC DEVICES SOLUTIONS COMPANY
* ALL RIGHTS RESERVED
*
* THE SOURCE CODE AND ITS RELATED DOCUMENTATION IS PROVIDED "AS IS". TOSHIBA
* CORPORATION MAKES NO OTHER WARRANTY OF ANY KIND, WHETHER EXPRESS, IMPLIED OR,
* STATUTORY AND DISCLAIMS ANY AND ALL IMPLIED WARRANTIES OF MERCHANTABILITY,
* SATISFACTORY QUALITY, NON INFRINGEMENT AND FITNESS FOR A PARTICULAR PURPOSE.
*
* THE SOURCE CODE AND DOCUMENTATION MAY INCLUDE ERRORS. TOSHIBA CORPORATION
* RESERVES THE RIGHT TO INCORPORATE MODIFICATIONS TO THE SOURCE CODE IN LATER
* REVISIONS OF IT, AND TO MAKE IMPROVEMENTS OR CHANGES IN THE DOCUMENTATION OR
* THE PRODUCTS OR TECHNOLOGIES DESCRIBED THEREIN AT ANY TIME.
*
* TOSHIBA CORPORATION SHALL NOT BE LIABLE FOR ANY DIRECT, INDIRECT OR
* CONSEQUENTIAL DAMAGE OR LIABILITY ARISING FROM YOUR USE OF THE SOURCE CODE OR
* ANY DOCUMENTATION, INCLUDING BUT NOT LIMITED TO, LOST REVENUES, DATA OR
* PROFITS, DAMAGES OF ANY SPECIAL, INCIDENTAL OR CONSEQUENTIAL NATURE, PUNITIVE
* DAMAGES, LOSS OF PROPERTY OR LOSS OF PROFITS ARISING OUT OF OR IN CONNECTION
* WITH THIS AGREEMENT, OR BEING UNUSABLE, EVEN IF ADVISED OF THE POSSIBILITY OR
* PROBABILITY OF SUCH DAMAGES AND WHETHER A CLAIM FOR SUCH DAMAGE IS BASED UPON
* WARRANTY, CONTRACT, TORT, NEGLIGENCE OR OTHERWISE.
*/

#include "ble_api_message.h"
#include "host_api.h"
#include "host_api_defs.h"
#include "uart1_function.h"
#include "debug.h"

/* Local definition */
#define UART1_BAUD_RATE (115200)
#define UART1_TP_CHAR_TIMER_US (8000)
#define UART_RING_BUFF_SIZE (512)

/* Local variable */
static unsigned char *rcvBuff;
static unsigned char *sndBuff;
static unsigned int rcvBuffSize;
static unsigned int rcvIntSize;
static unsigned int sndDataSize;
static unsigned int baudrate;

static struct bt_cmn_msg msg;

/*
* @brief Receive data via UART1
* @param[in] buff A pointer of receive data
* @param[in] size A size of receive data
*/
static void receive(unsigned char *buff, int size)
{
  int j;

  if (!size) {
    return;
  }

  uart1_receive_bytes_callback_isr(buff, size);

  uart1_ll_Receive_Next(rcvBuff, rcvBuffSize, rcvIntSize, size, &j);

  /* send an empty dummy message to our message loop to wake up main task */
  {
    msg.mbx_id = C2_MBOX_ID;
    msg.module_id = UART1_API_MOD_ID;
    msg.mem_flags = sizeof(struct bt_cmn_msg);
    msg.length = 0;

    OS_API_PostMessage(C2_MBOX_ID, (void *)&msg);
  }
}

/*
* @brief Callback function for receive complete
* @param buff A pointer of receive data
* @param size A size of receive data
*/
static void cbReceiveComplete(unsigned char *buff, int size)
{
  receive(buff, size);
}

/*
* @brief Callback for character timer
* @param buff A pointer of receive data
* @param size A size of receive data
*/
static void cbCharTimer(unsigned char *buff, int size)
{
  receive(buff, size);
  uart1_CharTimer_Enable();
}

/*
* @brief Callback for error
* @param buff   A pointer of receive data
* @param size   A size of receive data
* @param reason A reason of error
*/
static void cbError(unsigned char *buff, int size, UART_ERROR_REASON reason)
{
  /* Deny receive data. */
  uart1_ll_Receive_Start(rcvBuff, rcvBuffSize, rcvIntSize);
  uart1_CharTimer_Enable();
}

/*
* @brief Callback for completing to send
* @param buff A pointer of sending data
*/
static void cbSendComplete(unsigned char *buff)
{
  /* Release sndBuff */
  OS_API_ReleaseTcuMsgMemory(UART_EVT_MPF_ID, (SR_FAT_MSG *)buff);

  sndBuff = NULL;
  sndDataSize = 0;
}

/*
* @brief Callback for request to send
*/
static void cbSendRequest(void)
{
  /* Send sndBuff */
  uart1_Send_Start(sndBuff, sndDataSize);
}

/*
* @brief Send data via UART1
* @param buff A pointer of send data
* @param len  A length of send data
*/
void uart1_send(const uint8_t *buf, unsigned int len)
{
  unsigned int tmpLen = 0;
  int i = 0;

  tmpLen = len;

  while (tmpLen > 0) {
    /* Wait to finish previous sending */
    while (NULL != sndBuff) {
      OS_API_Sleep(10);
    }

    /* Get sending buffer */
    OS_API_GetTcuMsgMemory(UART_EVT_MPF_ID, (SR_FAT_MSG **)&sndBuff, CMN_NOWAIT);

    if (UART_EVT_MPF_ID_SIZE < tmpLen) {
      sndDataSize = UART_EVT_MPF_ID_SIZE;
      tmpLen -= UART_EVT_MPF_ID_SIZE;
    } else {
      sndDataSize = tmpLen;
      tmpLen = 0;
    }

    for (i = 0; i < sndDataSize; i++) {
      *(sndBuff + i) = *(buf + i);
    }

    uart1_Send_Req();
  }
}

/* Function definition */

/*
* @brief Initialize and open UART1
*/
void uart1_init(void)
{
  uart1_Initialize();

  /* Get receiving buffer  */
  OS_API_GetTcuMsgMemory(UART_ACL_CMD_MPF_ID, (SR_FAT_MSG **)&rcvBuff, CMN_NOWAIT);

  rcvBuffSize = UART_ACL_CMD_MPF_ID_SIZE;
  rcvIntSize = rcvBuffSize / 2;

  sndBuff = NULL;
  sndDataSize = 0;

  baudrate = UART1_BAUD_RATE;

  uart1_ll_Open(cbReceiveComplete, cbCharTimer, cbSendComplete, cbSendRequest, cbError, NULL, NULL, NULL);

  uart1_HwFlow_Disable();

  uart1_Set_BaudRate(baudrate);

  uart1_Set_CharTimer(UART1_TP_CHAR_TIMER_US);

  uart1_ll_Receive_Start(rcvBuff, rcvBuffSize, rcvIntSize);

  uart1_CharTimer_Enable();
}
